#include "game.h"
#include "util.h"
#include "conf.h"
#include "char.h"
#include "enemy.h"
#include "player.h"
#include "command.h"
#include "shot.h"
#include "xmltree.h"
#include "bulletml.h"
#include "timer.h"
#include "narihiraml.h"
#include "select.h"
#include "graphics.h"

#include <algorithm>
#include <functional>
#include <iostream>
#include <fstream>
#include <string>
#include <ctime>
#include <sstream>
#include <cassert>

#include <SDL.h>

int Game::turn() const {
	return timer_->turn();
}

double Game::fps() const {
	return timer_->fps();
}

int Game::title() {
	Select select(inputMgr_.get());
	if (select.selectBullet() != 0) return -1;

	bulletDir_ = select.bulletDir();

	narihira_.reset(new NarihiraML(
						"data/"+bulletDir_+"/"+bulletDir_+".xml"));

    window_->clear();
    window_->flip();

    return 0;
}

void Game::init () {
    const Conf* conf = Conf::instance();

    int ret = SDL_Init(SDL_INIT_VIDEO);
    ASSERT(ret == 0, "cannot init video\n");
    ret = SDL_Init(SDL_INIT_TIMER);
    ASSERT(ret == 0, "cannot init timer\n");

    SDL_WM_SetCaption("narihira", "narihira");

    srand(time(0));

    start_ = false;
    end_ = false;
	rank_ = conf->level() * 0.01;

    try {
		window_ = new MainWindow("bmp/black.bmp");
		window_->printInformation(std::cout);

		inputMgr_.reset(
			MSDL::generateInputFromConf("key.conf").release());

		player_.reset(new Player (Point (180, 460), *inputMgr_, WHITE));
		chars_.push_back(player_);
    }
    catch (...) {
		SDL_Quit();
    }
}

void Game::endEv() {
	SDL_Event ev;
	while (SDL_PollEvent(&ev)) {
		if (ev.type == SDL_QUIT) {
			end_ = true;
			quit_ = true;
		}
	}
}

void Game::reinit() {
    end_ = false;
    enemy_.clear();
    shot_.clear();
    chars_.clear();

    player_.reset(new Player(Point(180, 460), *inputMgr_, WHITE));
    chars_.push_back(player_);

    window_->clear();
}

void Game::eraseAllBullets() {
	for (Shots::iterator ites = shot_.begin(); ites != shot_.end(); ++ites) {
		(*ites)->die();
	}
	for (Enemys::iterator ite = enemy_.begin();
		 ite != enemy_.end(); ++ite)
	{
		if ((*ite)->isZako()) (*ite)->die();
	}
}

void Game::addShot(boost::shared_ptr<Shot> shot) {
	shot_.push_back(shot);
	chars_.push_back(shot);
}

void Game::addEnemy(boost::shared_ptr<Enemy> enemy) {
	enemy_.push_back(enemy);
	chars_.push_back(enemy);
}

int Game::start() {
    clearWindow();

    start_ = true;

	window_->source()->setAlpha(SDL_SRCALPHA|SDL_RLEACCEL, 128);

	timer_.reset(new Timer);

    while (!end_) {
		pass_ = false;
		timerEv();
		playerEv();
		enemyEv();
		shotEv();
		narihira_->event();
		updateWindow();
		clearWindow();
		deadEv();
		hitEv();
		endEv();
    }
	window_->source()->setAlpha(SDL_SRCALPHA|SDL_RLEACCEL, SDL_ALPHA_OPAQUE);
	clearWindow();

	int fps = (int)timer_->getTrueFps();
	timer_.reset(0);

	if (quit_) return -1;

	window_->putString(bulletDir_ + " ended.", Point(50, 50), Color::RED);
	window_->putString("you eat", Point(60, 90), Color::YELLOW);
	window_->putString(int2string(player_->eat(WHITE)) + " white bullets",
					   Point(60, 110), Color::WHITE);
	window_->putString(int2string(player_->eat(BLACK)) + " black bullets",
					   Point(60, 130), Color::BLACK);
	window_->putString("you died " + int2string(player_->dead()) + " times.",
					   Point(60, 160), Color::YELLOW);
	window_->putString(int2string(fps) + "FPS.", Point(300, 440), Color::BLUE);
	window_->flip();

    SDL_Delay(1000);
	window_->putString("push any key", Point(50, 190), Color::RED);
	window_->flip();

    start_ = false;

    if (waitAnyInput() == 0) {
		waitInputRelease();
		return 0;
	}

    return -1;
}

Game::Game()
    : bulletml_(0)
{}

Game::~Game() {
    delete window_;
    delete bulletml_;

    SDL_Quit();
}

void Game::timerEv() {
	timer_->waitEvent();
	for (Chars::iterator ite = chars_.begin(); ite != chars_.end(); ++ite) {
		(*ite)->countUp();
	}
}

void Game::playerEv() {
    player_->event();
}

void Game::enemyEv() {
	for (Enemys::iterator ite = enemy_.begin(); ite != enemy_.end(); ++ite) {
		(*ite)->huntEvent();
	}
}

void Game::shotEv() {
}

void Game::deadEv() {
    Chars::iterator ite;
    Shots::iterator ites;
    Enemys::iterator itee;

    for (ite = chars_.begin(); ite != chars_.end();) {
		Charactor* ch = ite->get();

		if (!ch->alive()) {
			for (ites = shot_.begin(); ites != shot_.end(); ++ites)
			{
				if (ites->get() == ch) {
					shot_.erase(ites);
					break;
				}
			}
			for (itee = enemy_.begin(); itee != enemy_.end(); ++itee)
			{
				if (itee->get() == ch) {
					enemy_.erase(itee);
					break;
				}
			}

			ite = chars_.erase(ite);
		}
		else ite++;
    }
}

void Game::hitEv() {
    Shots::iterator ites;
    for (ites = shot_.begin(); ites != shot_.end(); ites++) {
		Shot* sh = ites->get();
		if (player_->isHit(sh)) {
			sh->hit(player_.get());
			player_->hit(sh);
		}
    }
	Enemys::iterator itee;
    for (itee = enemy_.begin(); itee != enemy_.end(); itee++) {
		Enemy* en = itee->get();
		if (player_->isHit(en)) {
			en->hit(player_.get());
			player_->hit(en);
		}
    }
}

void Game::clearWindow() {
	if (!pass_) window_->clear();
}

void Game::updateWindow() {
	Chars::iterator ite;

    for (ite = chars_.begin(); ite != chars_.end(); ++ite) {
		(*ite)->moveValue();
	}
    for (ite = chars_.begin(); ite != chars_.end(); ++ite) {
		(*ite)->wallEvent();
	}
    if (!pass_) {
		for (ite = chars_.begin(); ite != chars_.end(); ++ite) {
			(*ite)->moveGraphic();
		}
	}

    /* // ꡼󥷥åȤȤ
	   if (turn_%20 == 0) {
	   std::ostringstream oss;
	   oss << "dump/" << turn_ << ".bmp";
	   window_->save(oss.str());
	   }
    */

    if (!pass_) window_->flip();
}

void Game::getEnemyPosAndSpd(std::vector<std::pair<Point, Point> >& out,
							 CharColor color) const
{
    Enemys::const_iterator itee;
    for (itee = enemy_.begin(); itee != enemy_.end(); itee++) {
		const Enemy* ene = itee->get();
		if (color != ene->color())
			out.push_back(std::make_pair(ene->center(), ene->spd()));
    }
    Shots::const_iterator ites;
    for (ites = shot_.begin(); ites != shot_.end(); ites++) {
		const Shot* shot = ites->get();
		if (color != shot->color())
			out.push_back(std::make_pair(shot->center(), shot->spd()));
    }
}

const Scroll* Game::scroll() const {
	return narihira_->scroll();
}

int Game::waitAnyInput() {
	do {
		SDL_Delay(50);
		SDL_PumpEvents();
		endEv();
		if (quit_) return -1;
		if (inputMgr_->getButton(3)) return -1;
	} while (!(inputMgr_->getButton(0) ||
			   inputMgr_->getButton(1)));

	return 0;
}

void Game::waitInputRelease() {
	while (inputMgr_->getAnyButton()) {
        SDL_Delay(50);
        SDL_PumpEvents();
    }
}



