#include "narihiraml.h"
#include "builtinfuncs.h"
#include "util.h"
#include "obstacle.h"
#include "battery.h"
#include "scroll.h"
#include "player.h"
#include "game.h"

#include <iostream>
#include <tinyxml.h>
#include <algorithm>
#include <functional>

NarihiraML::NarihiraML(const std::string& filename) {
	TiXmlDocument doc(filename);
	doc.LoadFile();

	std::cout << doc.Value() << ": parsing..." << std::endl;

	TiXmlNode* node;
	for (node = doc.FirstChild(); node; node = node->NextSibling()) {
		if (node->ToElement() != 0) {
			parseNarihira(node);
		}
	}
}

NarihiraML::~NarihiraML() {
	deleteClear(obstacles_);
	deleteClear(batteries_);
}

void NarihiraML::parseNarihira(TiXmlNode* node) {
	TiXmlNode* n;
	for (n = node->FirstChild(); n; n = n->NextSibling()) {
		if (n->Value() == "scroll") parseScroll(n);
		else if (n->Value() == "obstacle") parseObstacle(n);
		else if (n->Value() == "battery") parseBattery(n);
		else {
			ASSERT(0, "not narihira's child: " + n->Value());
		}
	}
}

void NarihiraML::parseScroll(TiXmlNode* node) {
	scroll_.reset(new Scroll);

	for (TiXmlNode* n = node->FirstChild(); n; n = n->NextSibling()) {
		if (n->Value() == "position") parsePosition(n);
		else if (n->Value() == "roll") parseRoll(n);
		else {
			ASSERT(0, "not scroll's child: " + n->Value());
		}
	}
}

void NarihiraML::parsePosition(TiXmlNode* node) {
	int num = 0;

	for (TiXmlNode* n = node->FirstChild(); n; n = n->NextSibling()) {
		ASSERT(n->Value() == "p", "not position's child: " + n->Value());
		TiXmlElement* e = n->ToElement();

		double x = str2double(*e->Attribute("x"));
		double y = str2double(*e->Attribute("y"));
		int t = str2int(*e->Attribute("t"));
		scroll_->setPosition(x, y, t);
		num++;
	}

	std::cout << "\t" << num << "positions found." << std::endl;
}

void NarihiraML::parseRoll(TiXmlNode* node) {
	int num = 0;

	for (TiXmlNode* n = node->FirstChild(); n; n = n->NextSibling()) {
		ASSERT(n->Value() == "r", "not roll's child: " + n->Value());
		TiXmlElement* e = n->ToElement();

		double d = str2double(*e->Attribute("d"));
		int t = str2int(*e->Attribute("t"));
		scroll_->setRoll(d, t);
		num++;
	}

	std::cout << "\t" << num << "rolls found." << std::endl;
}

void NarihiraML::parseObstacle(TiXmlNode* node) {
	int num = 0;

	for (TiXmlNode* n = node->FirstChild(); n; n = n->NextSibling()) {
		ASSERT(n->Value() == "o", "not obstacle's child: " + n->Value());
		TiXmlElement* e = n->ToElement();

		double x = str2double(*e->Attribute("x"));
		double y = str2double(*e->Attribute("y"));
		double w = str2double(*e->Attribute("w"));
		double h = str2double(*e->Attribute("h"));
		bool scroll = true;
		const std::string* scrollStr = e->Attribute("scroll");
		if (scrollStr != 0 && *scrollStr == "no") scroll = false;
		obstacles_.push_back(new Obstacle(Point(x, y), Point(w, h), scroll));
		num++;
	}

	std::cout << "\t" << num << "obstacles found." << std::endl;
}

void NarihiraML::parseBattery(TiXmlNode* node) {
	int num = 0;

	for (TiXmlNode* n = node->FirstChild(); n; n = n->NextSibling()) {
		ASSERT(n->Value() == "b", "not battery's child: " + n->Value());
		TiXmlElement* e = n->ToElement();

		double x = str2double(*e->Attribute("x"));
		double y = str2double(*e->Attribute("y"));
		const std::string& name = *e->Attribute("name");
		Battery* battery = new Battery(Point(x, y), name);

		const std::string* tmp;
		if ((tmp = e->Attribute("d")) != 0)
			battery->setDirection(str2double(*tmp));
		if ((tmp = e->Attribute("t")) != 0) battery->setTurn(str2int(*tmp));
		if ((tmp = e->Attribute("color")) != 0) {
			if (*tmp == "white") {
				battery->setColor(WHITE);
				battery->setChildColor(WHITE);
			}
			else if (*tmp == "black") {
				battery->setColor(BLACK);
				battery->setChildColor(BLACK);
			}
			else {
				int colorcode = str2int(*tmp);
				CharColor col = EMPTY;
				if (colorcode % 2 == 0) col = WHITE;
				else col = BLACK;
				battery->setChildColor(col);
				if (colorcode > 50) battery->setColor(EMPTY);
				else battery->setColor(col);
				int inv = (-colorcode+3)/2;
				battery->setInvisible(inv);
			}
		}
		if ((tmp = e->Attribute("rank")) != 0)
			battery->setRank(str2double(*tmp));
		if ((tmp = e->Attribute("laser")) != 0) {
			if (*tmp == "yes") battery->setIsLaser(true);
		}
		if ((tmp = e->Attribute("scroll")) != 0) {
			if (*tmp == "no") battery->setIsScroll(false);
		}
		if ((tmp = e->Attribute("width")) != 0)
			battery->setWidth(str2int(*tmp));
		else battery->setWidth(5);

		batteries_.push_back(battery);
		num++;
	}

	std::cout << "\t" << num << "batteries found." << std::endl;
}

void NarihiraML::event() {
	if (!Game::instance()->pass()) {
		drawObstacles();
	}
	batteryEvent();
	collisionObstacle();
	if (scroll_->isEnd()) Game::instance()->end();
}

void NarihiraML::drawObstacles() {
	std::for_each(obstacles_.begin(), obstacles_.end(),
				  std::mem_fun(&Obstacle::draw));
}

void NarihiraML::batteryEvent() {
	std::for_each(batteries_.begin(), batteries_.end(),
				  std::mem_fun(&Battery::event));
}

void NarihiraML::collisionObstacle() {
	Player* player = Game::instance()->player();
	if (std::find_if(obstacles_.begin(), obstacles_.end(),
					 std::bind2nd(std::mem_fun(&Obstacle::isHit), player))
		!= obstacles_.end())
	{
		player->die();
	}
}
