#ifndef BOHA_WIN
#define BOHA_WIN

#include "util.h"
#include "singleton.h"
#include "config.h"

#include <SDL.h>

#ifdef HAVE_SDL_SDL_TTF_H
#include <SDL_ttf.h>
#endif

#include <cassert>

#include <string>
#include <iostream>
#include <utility>
#include <memory>

// ɥ饹
class Window {
public:
	explicit Window(SDL_Surface* surf) : surf_(surf) {}
    explicit Window(const std::string& str);
	Window(const Window& rhs);
	Window(const Point& size, int bpp, Uint32 flag,
		   Uint32 Rmask =0, Uint32 Gmask =0, Uint32 Bmask =0, Uint32 Amask =0);
    Window(int x, int y, int bpp, Uint32 flag,
		   Uint32 Rmask =0, Uint32 Gmask =0, Uint32 Bmask =0, Uint32 Amask =0);

    virtual ~Window();

    void fillRect(const Rect& rect, int color) {
		int ret = SDL_FillRect(surf_, const_cast<SDL_Rect*>(rect.val()),
							   color);
		assert(ret == 0);
    }
	void fillRect(const Rect& rect, Color col) {
		fillRect(rect, mapRGB(col));
	}
	void fillRect(int color) {
		int ret = SDL_FillRect(surf_, 0, color);
		assert(ret == 0);
	}
	void fillRect(Color col) {
		fillRect(mapRGB(col));
	}

    void blit(Window* src) {
		int ret = SDL_BlitSurface(src->surf_, 0, surf_, 0);
		assert(ret == 0);
    }
    void blit(Window* src, const Rect& srcrect, const Rect& dstrect) {
		int ret =
			SDL_BlitSurface(src->surf_, const_cast<SDL_Rect*>(srcrect.val()),
							surf_, const_cast<SDL_Rect*>(dstrect.val()));
		assert(ret == 0);
    }
	void blit(Window* src, const Point& pnt) {
		SDL_Rect dst;
		dst.x = pnt.xi();
		dst.y = pnt.yi();
		int ret =
			SDL_BlitSurface(src->surf_, 0, surf_, &dst);
		assert(ret == 0);
	}

    void printInformation(std::ostream& os) const;

    virtual const Point& size() const { return size_; }

    void save(const std::string& filename) {
		SDL_SaveBMP(surf_, filename.c_str());
    }

	void setAlpha(Uint32 flag, Uint8 alpha) {
		int ret = SDL_SetAlpha(surf_, flag, alpha);
		assert(ret == 0);
	}

	void convertDisplayFormat();

    void setColorKey(Uint32 flag, Uint32 key) {
		int ret = SDL_SetColorKey(surf_, flag, key);
		assert(ret == 0);
    }
	void setColorKey(Uint32 flag, Color col) {
		setColorKey(flag, mapRGB(col));
	}

	void setMask(Color col);

	double halfDiagonal() const { return halfDiagonal_; }
	const Point& halfSize() const { return halfSize_; }

	void drawHLine(Sint16 x1, Sint16 x2, Sint16 y, Color col);
	void drawVLine(Sint16 x, Sint16 y1, Sint16 y2, Color col);
	void drawRect(Sint16 x1, Sint16 y1, Sint16 x2, Sint16 y2, Color col);
	void drawRect(const Rect& r, Color col) {
		drawRect(r.x(), r.y(), r.x2(), r.y2(), col);
	}

	void drawLine(const Point& pnt, const Point& spd,
				  int width, Color col1, Color col2);

protected:
	void lock() {
		SDL_LockSurface(surf_);
	}
	void unlock() {
		SDL_UnlockSurface(surf_);
	}

	Uint32 getPixel(int x, int y);
	void putPixel(int x, int y, Uint32 pixel);
	void putPixelHalf(int x, int y, Uint32 pixel);
    Uint32 mapRGB(Color col) const {
		return col.map(surf_->format);
    }
	Color getRGB(Uint32 pixel) const {
		Uint8 r, g, b;
		SDL_GetRGB(pixel, surf_->format, &r, &g, &b);
		return Color(r, g, b);
	}

protected:
	void calcCache();

protected:
	Window() {}

    SDL_Surface* surf_;
    Point size_;

	double halfDiagonal_;
	Point halfSize_;

};

// ᥤ󥦥ɥ
class MainWindow : public Window, public Singleton<MainWindow, CanNewPolicy> {
public:
    MainWindow(int x, int y, int bpp);
    MainWindow(const Point& size, int bpp);
    explicit MainWindow(const std::string& str);

	virtual ~MainWindow();

    void loadBMP(const std::string& str);
    void loadBMP(const std::string& str, const Point& pnt);

    void putString(const std::string& str, const Point& pnt,
				   const Color& col);

    virtual const Point& size() const { return first_->size(); }

    SDL_PixelFormat* format() { return surf_->format; }

    int flip() {
		return SDL_Flip(surf_);
    }

    void clear() {
		blit(first_.get());
    }
    void clearRect(const Rect& rect) {
		blit(first_.get(), rect, rect);
    }

	void reset() {
		blit(orig_.get());
		first_->blit(orig_.get());
	}

	Window* source() { return first_.get(); }

protected:
    static const Uint32 flag_;

    std::auto_ptr<Window> first_;
	std::auto_ptr<Window> orig_;

#ifdef HAVE_SDL_SDL_TTF_H
	TTF_Font* font_;
#else
    std::auto_ptr<Window> strings_;
    std::string sample_;
    static const double dotPerChar_;
    static const int charWidth_;
    static const int charHeight_;
#endif
};

#endif // BOHA_WIN
